
Require Export set1.

Set Implicit Arguments.
(* $Id: algebran.v,v 1.1 2010/08/02 13:05:01 grimm Exp $ *)



Module Operations.
Definition is_law1I (f:EE) (E:Set):=
  forall a, inc a E -> inc (f a) E.

Definition is_law2I (f2:EEE) (E:Set):=
  forall a b, inc a E -> inc b E -> inc (f2 a b) E.

Definition is_lawaI (f2:EEE) (A E:Set) :=
  (forall a b, inc a A -> inc b E -> inc (f2 a b) E).

Definition uncreate  (E:Set) (f:EE) :=
  L E (fun u=> f u).

Definition bincreate (E:Set) (f2:EEE) :=
  L (product E E) (fun u=> f2 (P u) (Q u)). 

Definition actcreate (A E:Set)(f2:EEE) :=
  L (product A E) (fun u=> f2(P u) (Q u)). 

Definition binapply (f a b: Set) := V (J a b) f.
Definition actapply (f a b: Set) := V (J a b) f.
Definition unapply (f a : Set) := V a f.

Definition is_law1E (f:EE)(E: Set):=
  forall a, ~ inc a E -> f a = emptyset.

Definition is_law2E (f2:EEE)(E:Set):=
  forall a b, ~ inc a E \/ ~ inc b E -> f2 a b = emptyset.

Definition is_lawaE (f2:EEE)(A E:Set):=
  forall a b, ~ inc a A \/ ~ inc b E -> f2 a b = emptyset.

Lemma unary_extensionality: forall f g E,
   is_law1E f E -> is_law1E g E ->
   (forall a, inc a E -> f a = g a) -> f = g.
Proof. ir. ap arrow_extensionality. ir.
  nin (inc_or_not a E). app H1. rww H. rww H0.
Qed.

Lemma binary_extensionality: forall f g E,
  is_law2E f E -> is_law2E g E ->
  (forall a b, inc a E -> inc b E -> f a b= g a b) -> f = g.
Proof. ir. ap arrow_extensionality. ir.
  ap arrow_extensionality. ir.  nin (inc_or_not a E). nin (inc_or_not a0 E).
  app H1.
  assert ( ~ inc a E \/ ~ inc a0 E). right. am. rww H; rww H0.
  assert ( ~ inc a E \/ ~ inc a0 E). left. am. rww H; rww H0.
Qed.

Lemma action_extensionality: forall f g A E,
  is_lawaE f A E-> is_lawaE g A E->
  (forall a b, inc a A -> inc b E ->f a b= g a b) -> f = g.
Proof. ir. ap arrow_extensionality. ir.
  ap arrow_extensionality. ir. nin (inc_or_not a A). nin (inc_or_not a0 E).
  app H1. 
  assert ( ~ inc a A \/ ~ inc a0 E). right. am. rww H; rww H0.
  ir. assert ( ~ inc a A \/ ~ inc a0 E). left. am. rww H; rww H0.
Qed.

Definition is_law1 (f:EE)(E:Set):=
   (is_law1I f E) & (is_law1E f E).

Definition is_law2 (f2:EEE)(E:Set):=
  (is_law2I f2 E) & (is_law2E f2 E).

Definition is_lawa (f2:EEE)(A E:Set):=
  (is_lawaI f2 A E) & (is_lawaE f2 A E).


Lemma bincreate_recovers: forall f2 E a b, 
  inc a E -> inc b E -> binapply (bincreate E f2) a b = f2 a b.
Proof. ir. uf binapply; uf bincreate. bw. aw. fprops.
Qed.

Lemma actcreate_recovers: forall f2 A E a b, 
  inc a A -> inc b E -> actapply (actcreate A E f2) a b = f2 a b.
Proof. ir. uf actapply; uf actcreate. bw.  aw. fprops.
Qed.

Lemma uncreate_recovers: forall f E a, 
  inc a E -> unapply (uncreate E f) a  = f a.
Proof. ir. uf unapply. bw.  
Qed.

Definition binapply_x (f E a b: Set) := 
  Yo (inc a E) (Yo (inc b E) (V (J a b) f) emptyset) emptyset.
Definition unapply_x  (f E a : Set) := Yo (inc a E) (V a f) emptyset.
Definition actapply_x (f A E a b: Set) := 
  Yo (inc a A) (Yo (inc b E) (V (J a b) f) emptyset) emptyset.


Lemma is_law1E_unapply0: forall f E a, 
  ~ inc a E -> unapply_x f E a = emptyset.
Proof. ir. uf unapply_x. rww Y_if_not_rw.
Qed.

Lemma is_law1E_unapply: forall f E, 
  is_law1E (unapply_x f E) E.
Proof. ir. red. ir. app is_law1E_unapply0.
Qed.

Lemma is_law1E_unapply2: forall f E, 
  is_law1E (fun a  => unapply_x f E a) E.
Proof. ir. red. ir. app is_law1E_unapply0.
Qed.

Lemma is_law2E_binapply0: forall f E a b,
  ~ inc a E \/ ~ inc b E -> binapply_x f E a b = emptyset.
Proof. ir. unfold binapply_x. nin (inc_or_not a E). app Y_if.
  assert (~ inc b E).  induction H.  elim (H H0). am.
  rww Y_if_not_rw. rww Y_if_not_rw.  
Qed. 

Lemma is_law2E_binapply: forall f E, 
  is_law2E (binapply_x f E) E.
Proof. ir. red. ir. app is_law2E_binapply0.
Qed. 

Lemma is_law2E_binapply2:  forall f E, 
  is_law2E (fun a b  => binapply_x f E a b) E.
Proof. ir. red. ir.  app is_law2E_binapply0.
Qed.

Lemma is_lawaE_actapply0: forall f A E a b,
  ~ inc a A \/ ~ inc b E -> actapply_x f A E a b = emptyset.
Proof. ir. uf actapply_x. nin (inc_or_not a A). app Y_if.
  assert (~ inc b E).  induction H.  elim (H H0). am.
  rww Y_if_not_rw. rww Y_if_not_rw.  
Qed. 

Lemma is_lawaE_actapply: forall f A E, 
  is_lawaE (actapply_x f A E) A E.
Proof. ir. red. ir. app is_lawaE_actapply0.
Qed. 

Lemma is_lawaE_actapply2: forall f A E,
  is_lawaE (fun a b => actapply_x f A E a b) A E.
Proof. ir. red. ir. app is_lawaE_actapply0.
Qed.


Lemma unary_recovers: forall f E a, 
  inc a E -> unapply_x (uncreate E f) E a = f a.
Proof. ir. uf unapply_x. app Y_if. uf uncreate.  bw. 
Qed.

Lemma binary_recovers: forall f2 E a b, 
  inc a E -> inc b E -> binapply_x (bincreate E f2) E a b = f2 a b.
Proof. ir. uf binapply_x. app Y_if. app Y_if.
  rewrite <- bincreate_recovers with (E:=E). uf binapply. tv. am. am.
Qed.

Lemma action_recovers: forall f2 A E a b, 
  inc a A -> inc b E -> actapply_x (actcreate A E f2) A E a b = f2 a b.
Proof. ir. uf actapply_x. app Y_if. app Y_if.
  rewrite <- actcreate_recovers with (A:=A)(E:=E). uf actapply. tv. am. am.
Qed.

Definition unary_graph (f E:Set) :=
   (fgraph f) & (domain f = E) & (sub (range f) E).

Definition binary_graph (f E:Set) :=
  (fgraph f) &  (domain f = product E E)  & (sub (range f) E).

Definition action_graph (f A E:Set) :=
  (fgraph f) &  (domain f = product A E) & (sub (range f) E).

Lemma unary_from_arrow: forall E f, 
  is_law1I f E -> unary_graph (L E f) E. 
Proof. ir. red. assert(fgraph (L E f)).  gprops.
  assert (domain (L E f) = E). bw. 
  split. am. split. am. red. ir.
  rwi frange_inc_rw H2. rwi H1 H2. nin H2. nin H2. rw H3. bw. app H. am.
Qed.

Lemma binary_from_arrow: forall E f2,
 is_law2I f2 E->  binary_graph (bincreate E f2) E. 
Proof. ir. uf bincreate. red. split. gprops. split.  bw.
  red. ir. rwi frange_inc_rw H0.  nin H0. nin H0. bwi H0.  bwi H1. rw H1.
  awi H0.  ee. app H. am. gprops.
Qed.

Lemma action_from_arrow: forall A B f2, 
 is_lawaI f2 A B ->  action_graph (actcreate A B f2) A B. 
Proof. ir. uf actcreate. red.  bw.  split. gprops. split. tv. 
  red. ir. rwi frange_inc_rw H0. nin H0. nin H0. bwi H0. bwi H1. rw H1.
  awi H0. ee. app H.  am. gprops.
Qed.

End Operations.

(** **  Notations *)

Module Notation.

Require Import String.

Section Definitions.
Open Local Scope string_scope. 

Definition Plus_not := Ro "Plus".
Definition Zero_not := Ro "Zero".
Definition Neg_not := Ro "Neg".
Definition Mult_not := Ro "Mult".
Definition One_not := Ro "One".
Definition Inv_not := Ro "Inv".
Definition Laction_not := Ro "Laction".
Definition Raction_not := Ro "Raction".
Definition Laux_not := Ro "Lauxiliary".
Definition Raux_not := Ro "Rauxiliary".


Definition Underlying_not := Ro "Underlying".  
Definition Source_not  := Ro "Source". 
Definition Target_not  := Ro "Target". 
Definition Arrow_not  := Ro "Arrow". 

End Definitions.

Definition sourceC x := V Source_not x.
Definition targetC x := V Target_not x.
Definition arrowC x := V Arrow_not x.

Definition Ul x  := V Underlying_not x.
Definition sourceU x := Ul (sourceC x). 
Definition targetU x := Ul (targetC x). 
Definition ev u x := V x (arrowC u).


Definition lauxiliary x  := (V Laux_not x).
Definition rauxiliary x := (V Raux_not x).
Definition dplus x a b:=  Operations.binapply_x (V Plus_not x) (Ul x) a b.
Definition zero x := (V Zero_not x).
Definition neg x a := Operations.unapply_x (V Neg_not x) (Ul x) a.
Definition dmult x a b :=  Operations.binapply_x (V Mult_not x) (Ul x) a b.
Definition one x  := (V One_not x).
Definition inv x a  := Operations.unapply_x (V Inv_not x) (Ul x) a.
Definition lmult x a b  :=  
  Operations.actapply_x (V Laction_not x) (Ul (lauxiliary x)) (Ul x) a b.
Definition rmult x a b :=  
  Operations.actapply_x (V Raction_not x) (Ul (rauxiliary x)) (Ul x) a b.


Lemma is_bin_domain_plus: forall x w, Ul w = x ->
  Operations.is_law2E (dplus w) x.
Proof. ir. unfold dplus. rw H. ap Operations.is_law2E_binapply2.
Qed.

Lemma is_bin_domain_mult: forall x w, Ul w = x ->
  Operations.is_law2E (dmult w) x.
Proof. ir. unfold dmult. rw H. ap Operations.is_law2E_binapply2.
Qed.

Lemma is_un_domain_neg: forall x w, Ul w = x ->
  Operations.is_law1E (neg w) x.
Proof. ir. uf neg. rw H. apply Operations.is_law1E_unapply2.
Qed.

Lemma is_un_domain_inv: forall x w, Ul w = x ->
  Operations.is_law1E (inv w) x.
Proof. ir. uf inv. rw H. apply Operations.is_law1E_unapply2.
Qed.

Lemma is_act_domain_lmult: forall x y w, Ul w = x ->
  lauxiliary w = y ->
  Operations.is_lawaE (lmult w) (Ul y) x.
Proof. ir. unfold lmult. rw H. rw H0. ap Operations.is_lawaE_actapply2.
Qed.

Lemma is_act_domain_rmult: forall x y w, Ul w = x ->
  rauxiliary w =y ->
  Operations.is_lawaE (rmult w) (Ul y) x.
Proof. ir. unfold rmult. rw H. rw H0. ap Operations.is_lawaE_actapply2.
Qed.

(** A notation is a function taking a string as argument *)

Definition is_notation f :=
  fgraph f & domain f = string.

(** A special notaion is [stop] indicating the end of the list *)
Definition stop := L string (fun s => emptyset).

Lemma is_notation_stop : is_notation stop.
Proof. red. uf stop. split. gprops. bw.
Qed.

Lemma V_stop : forall x, V x stop = emptyset.
Proof. ir. uf stop. 
  apply by_cases with (inc x string); intros. bw. app L_V_out.
Qed.

(** We use [denote] to add a pair to a notation *)

Definition denote str obj old :=
  L string (fun s => (Yo (s = str) obj (V s old))).

Lemma is_notation_denote : forall str obj old,
  is_notation old -> is_notation (denote str obj old).
Proof. ir. red. unfold denote. split.  gprops. bw.
Qed.

Lemma V_denote_new : forall str obj old x,
  x = str -> inc x string -> V x (denote str obj old) = obj.
Proof. ir. uf denote. bw. rww Y_if_rw.
Qed.

Lemma V_denote_old : forall str obj old x,
  ~x=str -> inc x string -> V x (denote str obj old) = V x old.
Proof. ir. uf denote. bw. rww Y_if_not_rw.
Qed. 

(** We use injectivity of [Ro] on strings in the tactics that follow *)
Lemma R_string_inj : forall a b:string,
  ~a=b -> ~(Ro a = Ro b).
Proof. ir. red. ir. ap H. app R_inj. 
Qed.


Ltac show_string :=
  match goal with
    | |- inc ?X1 string  => try assumption; unfold X1; apply R_inc
    | _ => fail
  end.
 
Ltac show_different :=
  match goal with
    | |- ~(?X1 = ?X2) => 
      (try (unfold X1); try (unfold X2);
        apply R_string_inj; discriminate)
    | _ => fail
  end. 

Ltac drw_solve := solve [trivial | show_different | assumption | show_string].

Ltac drw_new := 
  rewrite V_denote_new; [idtac | drw_solve | drw_solve].

Ltac drw_old :=
  rewrite V_denote_old; [idtac | drw_solve | drw_solve].

Ltac drw_stop := rewrite V_stop.

Ltac drw := first [
  drw_new ; drw | drw_old; drw | drw_stop ; drw | 
    trivial | assumption | idtac].


Ltac drwe := first [ drw_new ; drwe | drw_old; drwe| idtac].
Ltac mw:= autorewrite with mw; try tv; try am.

End Notation. 
Export Notation. 
