Require Export set2.
Set Implicit Arguments.
Unset Strict Implicit.

(* $Id: group.v,v 1.18 2009/11/09 11:53:22 grimm Exp $ *)

Module Operations. 

Definition is_unary_function (f:EE)(x:Bset):=
  (forall a, inc a x -> inc (f a) x).

Definition is_binary_function (f2:EEE)(x:Bset):=
  (forall a b, inc a x -> inc b x -> inc (f2 a b) x).

Definition is_action_function(f2:EEE)(x y:Bset) :=
  (forall a b, inc a x -> inc b y -> inc (f2 a b) y).

Definition is_unary_domain (f:EE)(x: Bset):=
  forall a, ~ inc a x -> f a = emptyset.

Definition is_binary_domain (f2:EEE)(x:Bset):=
  forall a b, ~ inc a x \/ ~ inc b x -> f2 a b = emptyset.

Definition is_action_domain (f2:EEE)(x y:Bset):=
  forall a b, ~ inc a x \/ ~ inc b y -> f2 a b = emptyset.

Definition is_strong_unary_function (f:EE)(x:Bset):=
   (is_unary_function f x) & (is_unary_domain f x).

Definition is_strong_binary_function (f2:EEE)(x:Bset):=
  (is_binary_function f2 x) & (is_binary_domain f2 x).

Definition is_strong_action_function (f2:EEE)(x y:Bset):=
  (is_action_function f2 x y) & (is_action_domain f2 x y).


Lemma unary_extensionality: forall f g x,
   is_unary_domain f x-> is_unary_domain g x ->
   (forall a, inc a x -> f a = g a) -> f = g.
Proof. ir. ap arrow_extensionality. ir.
  nin (inc_or_not a x). app H1. rww H. rww H0.
Qed.

Lemma binary_extensionality: forall f g x,
      is_binary_domain f x-> is_binary_domain g x ->
   (forall a b, inc a x -> inc b x ->f a b= g a b) -> f = g.
Proof. ir. ap arrow_extensionality. ir.
  ap arrow_extensionality. ir.  nin (inc_or_not a x). nin (inc_or_not a0 x).
  app H1.
  assert ( ~ inc a x \/ ~ inc a0 x). right. am. rww H; rww H0.
  assert ( ~ inc a x \/ ~ inc a0 x). left. am. rww H; rww H0.
Qed.

Lemma action_extensionality: forall f g x y,
      is_action_domain f x y-> is_action_domain g x y->
   (forall a b, inc a x -> inc b y ->f a b= g a b) -> f = g.
Proof. ir. ap arrow_extensionality. ir.
  ap arrow_extensionality. ir. nin (inc_or_not a x). nin (inc_or_not a0 y).
  app H1. 
  assert ( ~ inc a x \/ ~ inc a0 y). right. am. rww H; rww H0.
  ir. assert ( ~ inc a x \/ ~ inc a0 y). left. am. rww H; rww H0.
Qed.


Definition bincreate (x:Bset)(f2:EEE) :=
  L (product x x) (fun u=> f2 (P u) (Q u)). 

Definition actcreate (x y:Bset)(f2:EEE) :=
  L (product x y) (fun u=> f2(P u) (Q u)). 

Definition binapply (f a b: Bset) := V (J a b) f.
Definition actapply (f a b: Bset) := V (J a b) f.

Lemma bincreate_recovers: forall f2 x a b, 
  inc a x -> inc b x -> binapply (bincreate x f2) a b = f2 a b.
Proof. ir. uf binapply; uf bincreate. bw. aw. fprops.
Qed.

Lemma actcreate_recovers: forall f2 x y a b, 
  inc a x -> inc b y -> actapply (actcreate x y f2) a b = f2 a b.
Proof. ir. uf actapply; uf actcreate. bw.  aw. fprops.
Qed.

Definition binapply_x (f x a b: Bset) := 
  Yo (inc a x) (Yo (inc b x) (V (J a b) f) emptyset) emptyset.
Definition unapply_x  (f x a : Bset) := Yo (inc a x) (V a f) emptyset.
Definition actapply_x (f x y a b: Bset) := 
  Yo (inc a x) (Yo (inc b y) (V (J a b) f) emptyset) emptyset.

Lemma is_unary_domain_unapply0: forall f x a, 
  ~ inc a x -> unapply_x f x a = emptyset.
Proof. ir. uf unapply_x. rww Y_if_not_rw.
Qed.

Lemma is_unary_domain_unapply: forall f x, 
  is_unary_domain (unapply_x f x) x.
Proof. ir. red. ir. app is_unary_domain_unapply0.
Qed.

Lemma is_unary_domain_unapply2: forall f x, 
  is_unary_domain (fun a  => unapply_x f x a) x.
Proof. ir. red. ir. app is_unary_domain_unapply0.
Qed.

Lemma is_binary_domain_binapply0: forall f x a b,
  ~ inc a x \/ ~ inc b x -> binapply_x f x a b = emptyset.
Proof. ir. unfold binapply_x. nin (inc_or_not a x). app Y_if.
  assert (~ inc b x).  induction H.  elim (H H0). am.
  rww Y_if_not_rw. rww Y_if_not_rw.  
Qed. 

Lemma is_binary_domain_binapply: forall f x, 
  is_binary_domain (binapply_x f x) x.
Proof. ir. red. ir. app is_binary_domain_binapply0.
Qed. 

Lemma is_binary_domain_binapply2:  forall f x, 
  is_binary_domain (fun a b  => binapply_x f x a b) x.
Proof. ir. red. ir.  app is_binary_domain_binapply0.
Qed.

Lemma is_action_domain_actapply0: forall f x y a b,
  ~ inc a x \/ ~ inc b y -> actapply_x f x y a b = emptyset.
Proof. ir. uf actapply_x. nin (inc_or_not a x). app Y_if.
  assert (~ inc b y).  induction H.  elim (H H0). am.
  rww Y_if_not_rw. rww Y_if_not_rw.  
Qed. 

Lemma is_action_domain_actapply: forall f x y, 
  is_action_domain (actapply_x f x y) x y.
Proof. ir. red. ir. app is_action_domain_actapply0.
Qed. 

Lemma is_action_domain_actapply2: forall f x y,
  is_action_domain (fun a b => actapply_x f x y a b) x y.
Proof. ir. red. ir. app is_action_domain_actapply0.
Qed.

Lemma unary_recovers: forall f x a, 
  inc a x -> unapply_x (unary x f) x a = f a.
Proof. ir. uf unapply_x. app Y_if. uf unary.  bw. 
Qed.

Lemma binary_recovers: forall f2 x a b, 
  inc a x -> inc b x -> binapply_x (bincreate x f2) x a b = f2 a b.
Proof. ir. uf binapply_x. app Y_if. app Y_if.
  rewrite <- bincreate_recovers with (x:=x). uf binapply. tv. am. am.
Qed.

Lemma action_recovers: forall f2 x y a b, 
  inc a x -> inc b y -> actapply_x (actcreate x y f2) x y a b = f2 a b.
Proof. ir. uf actapply_x. app Y_if. app Y_if.
  rewrite <- actcreate_recovers with (x:=x)(y:=y). uf actapply. tv. am. am.
Qed.

Definition unary_operation (f x:Bset) :=
   (fgraph f) & (domain f = x) & (sub (range f) x).

Definition binary_operation (f x:Bset) :=
  (fgraph f) &  (domain f = Cartesian.product x x)  & (sub (range f) x).

Definition action_operation (f x y:Bset) :=
  (fgraph f) &  (domain f = product x y) & (sub (range f) y).


(* Montre que L x f est un operation unaire *)
Lemma unary_from_arrow: forall x f, 
  is_unary_function f x -> unary_operation (L x f) x. 
Proof. ir. red. assert(fgraph (L x f)).  gprops.
  assert (domain (L x f) = x). bw. 
  split. am. split. am. red. ir.
  rwi frange_inc_rw H2. rwi H1 H2. nin H2. nin H2. rw H3. bw. app H. am.
Qed.

(* Montre que f qui vient de f2 est une operation binaire  *)
Lemma binary_from_arrow: forall x f2,
 is_binary_function f2 x->  binary_operation (bincreate x f2) x. 
Proof. ir. uf bincreate. red. split. gprops. split.  bw.
  red. ir. rwi frange_inc_rw H0.  nin H0. nin H0. bwi H0.  bwi H1. rw H1.
  awi H0.  ee. app H. am. gprops.
Qed.


(* Montre que f qui vient de f2 est une action  *)
Lemma action_from_arrow: forall x y f2, 
 is_action_function f2 x y->  action_operation (actcreate x y f2) x y. 
Proof. ir. uf actcreate. red.  bw.  split. gprops. split. tv. 
  red. ir. rwi frange_inc_rw H0. nin H0. nin H0. bwi H0. bwi H1. rw H1.
  awi H0. ee. app H.  am. gprops.
Qed.

End Operations.


(* ******************************************************************* *)


(* the trivial tactic does not work in some cases *)
Ltac drwe := first [ drw_new ; drwe | drw_old; drwe| idtac].

Ltac mw:= autorewrite with mw; try tv; try am.

Module Monoid_notation. (* monoid *)

Open Local Scope string_scope.

Definition Plus_not := Ro "Plus".
Definition Zero_not := Ro "Zero".
Definition Neg_not := Ro "Neg".
Definition Mult_not := Ro "Mult".
Definition One_not := Ro "One".
Definition Inv_not := Ro "Inv".
Definition Laction_not := Ro "Laction".
Definition Raction_not := Ro "Raction".
Definition Laux_not := Ro "Lauxiliary".
Definition Raux_not := Ro "Rauxiliary".

Definition lauxiliary  (x : Bset) := (V Laux_not x).
Definition rauxiliary  (x : Bset) := (V Raux_not x).
Definition dplus (x a b : Bset) :=  Operations.binapply_x (V Plus_not x) (Ul x) a b.
Definition zero  (x : Bset) := (V Zero_not x).
Definition neg (x a : Bset) := Operations.unapply_x (V Neg_not x) (Ul x) a.
Definition dmult (x a b : Bset) :=  Operations.binapply_x (V Mult_not x) (Ul x) a b.
Definition one (x : Bset) := (V One_not x).
Definition inv (x a : Bset) := Operations.unapply_x (V Inv_not x) (Ul x) a.
Definition lmult (x a b : Bset) :=  
  Operations.actapply_x (V Laction_not x) (Ul (lauxiliary x)) (Ul x) a b.
Definition rmult (x a b : Bset) :=  
  Operations.actapply_x (V Raction_not x) (Ul (rauxiliary x)) (Ul x) a b.


Lemma is_bin_domain_plus: forall x w, Ul w = x ->
  Operations.is_binary_domain (dplus w) x.
Proof. ir. unfold dplus. rw H. ap Operations.is_binary_domain_binapply2.
Qed.

Lemma is_bin_domain_mult: forall x w, Ul w = x ->
  Operations.is_binary_domain (dmult w) x.
Proof. ir. unfold dmult. rw H. ap Operations.is_binary_domain_binapply2.
Qed.

Lemma is_un_domain_neg: forall x w, Ul w = x ->
  Operations.is_unary_domain (neg w) x.
Proof. ir. uf neg. rw H. apply Operations.is_unary_domain_unapply2.
Qed.

Lemma is_un_domain_inv: forall x w, Ul w = x ->
  Operations.is_unary_domain (inv w) x.
Proof. ir. uf inv. rw H. apply Operations.is_unary_domain_unapply2.
Qed.

Lemma is_act_domain_lmult: forall x y w, Ul w = x ->
  lauxiliary w = y ->
  Operations.is_action_domain (lmult w) (Ul y) x.
Proof. ir. unfold lmult. rw H. rw H0. ap Operations.is_action_domain_actapply2.
Qed.

Lemma is_act_domain_rmult: forall x y w, Ul w = x ->
  rauxiliary w =y ->
  Operations.is_action_domain (rmult w) (Ul y) x.
Proof. ir. unfold rmult. rw H. rw H0. ap Operations.is_action_domain_actapply2.
Qed.


Section Gconstruction. 
Variables (x:Bset) (p: EEE)(z:Bset) .

Definition create :=
  denote Underlying x 
  (denote Plus_not (Operations.bincreate x p)
      (denote Zero_not z stop)).

Lemma underlying_create: Ul create = x.
Proof. ir. uf Ul. uf create. drw.
Qed.

Lemma plus_create_when: forall a b, inc a x -> inc b x -> 
  dplus create a b = (p a b).
Proof. ir. uf dplus. uf create. uf Ul. drw. app Operations.binary_recovers.
Qed.

Lemma zero_create: zero create  = z.
Proof. uf zero. uf create. drw.
Qed.

Definition struct_property:=
  (inc z x) &
  (Operations.is_binary_function p x).

Lemma inc_create_zero_U: struct_property -> inc (zero create) (Ul create).
Proof. ir. rww zero_create. rww underlying_create. destruct H. am.
Qed.

Lemma is_bin_plus_create : struct_property
  -> Operations.is_strong_binary_function (dplus create) x.
Proof. ir. destruct H as [_ H]. red in H. red. split. 
   red. intros. rww plus_create_when. app H.  
   pose (H1:=underlying_create). app is_bin_domain_plus.
Qed.

Lemma plus_create : 
  Operations.is_binary_domain p x  ->  struct_property -> dplus create = p.
Proof. ir. apply Operations.binary_extensionality with x.
   pose (is_bin_plus_create H0). destruct i. am. am.
   app plus_create_when.
Qed.

End Gconstruction.

Hint Rewrite underlying_create plus_create plus_create_when zero_create: mw.

Lemma create_properties: forall x p z,
    Operations.is_binary_domain p x ->
    struct_property x p z ->  (dplus (create x p z) = p).
Proof. ir. ee. mw.  
Qed.

Definition like (a:Bset):= a = create (Ul a) (dplus a) (zero a).

Lemma create_like: forall a, like(create (Ul a) (dplus a) (zero a)).
Proof. ir. uf like.
  set (x:=(create (Ul a) (dplus a) (zero a))).
  assert (Ul x = Ul a). uf x; mw.
  assert (zero x = zero a). uf x; mw.
  assert (dplus x = dplus a). unfold x. 
  apply Operations.binary_extensionality with (Ul x).  rw H.
  app is_bin_domain_plus. app is_bin_domain_plus. sy; am.
  ir. rewrite H in H1,H2. mw.
  rw H; rw H0; rw H1. tv.
Qed.

Lemma like_create:forall x p z,
  Operations.is_binary_domain p x -> struct_property x p z ->
  like (create x p z).
Proof. ir. pose (create_properties H H0). ee. red. mw. 
Qed.

End Monoid_notation.
Export Monoid_notation.

Module Gnotation.

Section Gconstruction. (* Group *)
Variables (x:Bset) (p: EEE)(o: EE)(z:Bset) .

Definition create := 
  (denote Neg_not (unary x o) (Monoid_notation.create x p z)).

Lemma underlying_create: Ul create = x.
Proof. ir. uf Ul. uf create. unfold Monoid_notation.create. drw.
Qed.

Lemma plus_create_when: forall a b, inc a x -> inc b x -> 
  dplus create a b = (p a b).
Proof. ir. uf dplus.  uf Ul. uf create. unfold Monoid_notation.create. drw.  
  app Operations.binary_recovers.
Qed.

Lemma zero_create: zero create  = z.
Proof. uf zero. uf create. unfold Monoid_notation.create. drw.
Qed.

Lemma neg_create_when: forall a, inc a x -> neg create a = o a.
Proof. ir. uf neg. uf create. uf Ul. unfold Monoid_notation.create. drw. 
  app Operations.unary_recovers.
Qed.

Definition struct_property:=
  (inc z x) &
  (Operations.is_unary_function o x) &
  (Operations.is_binary_function p x).

Lemma inc_create_zero_U: struct_property -> inc (zero create) (Ul create).
Proof. ir. rw underlying_create; rww zero_create. destruct H. am.
Qed.

Lemma is_bin_plus_create : struct_property
  -> Operations.is_strong_binary_function (dplus create) x.
Proof. ir. destruct H as [_ [_ H]]. red in H. red. split. 
   red. intros. rww plus_create_when. app H.  
   pose (H1:=underlying_create). app is_bin_domain_plus.
Qed.

Lemma is_un_neg_create : struct_property
  -> Operations.is_strong_unary_function (neg create) x.
Proof. intros. destruct H as [_ [H _]]. 
  red in H. red. split. 
  red. intros. rww neg_create_when. app H. 
  pose (H1:=underlying_create). app is_un_domain_neg.
Qed.

Lemma plus_create : 
  Operations.is_binary_domain p x  ->  struct_property -> dplus create = p.
Proof. ir. apply Operations.binary_extensionality with x.
   pose (is_bin_plus_create H0). destruct i. am. am.
   app plus_create_when.
Qed.

Lemma neg_create : 
  Operations.is_unary_domain o x -> struct_property -> neg create = o.
Proof. intros. apply Operations.unary_extensionality with x.
  pose (is_un_neg_create H0). destruct i.  am. am. 
  intros. app neg_create_when.
Qed.

End Gconstruction.

Hint Rewrite underlying_create neg_create plus_create zero_create: mw.
Hint Rewrite neg_create_when plus_create_when: mw.
Lemma create_properties:  forall x p o z,
    Operations.is_unary_domain o x -> Operations.is_binary_domain p x ->
    struct_property x p o z ->
    ( (dplus (create x p o z) = p) & (neg (create x p o z) = o) ).
Proof. ir. ee. mw.  mw.  
Qed.

Definition like (a:Bset):= a = create (Ul a) (dplus a) (neg a) (zero a).

Lemma create_like: forall a, like(create (Ul a) (dplus a) (neg a) (zero a)).
Proof. ir. uf like.
  set (x:=(create (Ul a) (dplus a) (neg a) (zero a))).
  assert (Ul x = Ul a). uf x; mw.
  assert (zero x = zero a). uf x; mw.
  assert (neg x = neg a). uf x.
  apply Operations.unary_extensionality with (Ul x).  rw H.
  app is_un_domain_neg. app is_un_domain_neg. sy; am. ir. rwi H H1. mw.
  assert (dplus x = dplus a). unfold x. 
  apply Operations.binary_extensionality with (Ul x).  rw H.
  app is_bin_domain_plus. app is_bin_domain_plus. sy; am.
  ir. rewrite H in H2,H3. mw.
  rw H; rw H0; rw H1; rw H2. tv.
Qed.

Lemma like_create:forall x p o z,
    Operations.is_unary_domain o x -> Operations.is_binary_domain p x ->
    struct_property x p o z ->
    like (create x p o z).
Proof. ir. pose (create_properties H H0 H1). ee.
  red. rw H2; rw H3. mw. 
Qed.

End Gnotation.
Export Gnotation.

(* ************************************************** *)

Module Rnotation. (* Ring construction *)
Section Dconstruction.
  Variables (x:Bset) (p:EEE)(o: EE) (z:Bset)(u:Bset)(m:EEE).

Definition create :=
  (denote Mult_not (Operations.bincreate x m)
    (denote One_not u (Gnotation.create x p o z))).

Lemma underlying_create: Ul create = x.
Proof. ir. transitivity (Ul (Gnotation.create x p o z)).
  uf Ul; uf create. drwe. tv. mw.
Qed.

Lemma zero_create: zero create  = z.
Proof. ir. transitivity (zero (Gnotation.create x p o z)). 
  uf zero; uf create. drw. mw.
Qed.

Lemma neg_create_when: forall a, inc a x -> neg create a = o a.
Proof. ir. transitivity (neg (Gnotation.create x p o z) a).
  uf create; uf neg; uf Ul. drwe.  tv.  mw.  
Qed.

Lemma plus_create_when: forall a b, inc a x -> inc b x -> 
  dplus create a b = (p a b).
Proof. ir. transitivity (dplus (Gnotation.create x p o z) a b).
  uf create; uf dplus; uf Ul. drwe. tv. mw.
Qed.

Lemma mult_create_when: forall a b, inc a x -> inc b x -> 
  dmult create a b = (m a b).
Proof. ir. pose (underlying_create). ufi create e.
  uf dmult; uf create. rw e. drwe. app Operations.binary_recovers.
Qed.

Lemma one_create: one create = u.
Proof. uf one. uf create. drw.
Qed.

Definition struct_property:=
  (inc u x) &
  (Operations.is_binary_function m x) &
  (Gnotation.struct_property x p o z).

Lemma inc_create_zero_U: struct_property -> inc (zero create) (Ul create).
Proof. ir. rw zero_create. rw underlying_create. 
  destruct H. ee. destruct H1. am.
Qed.

Lemma is_bin_plus_create : struct_property
  -> Operations.is_strong_binary_function (dplus create) x.
Proof. ir. destruct H as  [_ [_ H]]. 
  destruct H as [_ [_ H]]. red in H. red. split. 
  red. intros. rww plus_create_when. app H.  
  pose (H1:=underlying_create). app is_bin_domain_plus.
Qed.

Lemma is_bin_mult_create : struct_property
  -> Operations.is_strong_binary_function (dmult create) x.
Proof. ir. destruct H as [_ [H _]]. red in H. red. split. 
   red. intros. rww mult_create_when. app H.  
   pose (H1:=underlying_create). app is_bin_domain_mult.
Qed.

Lemma is_un_neg_create : struct_property
  -> Operations.is_strong_unary_function (neg create) x.
Proof. intros. destruct H as [_ [_ H]]. destruct H as [_ [H _]]. 
  red in H. red. split. 
  red. intros. rww neg_create_when. app H. 
  pose (H1:=underlying_create). app is_un_domain_neg.
Qed.

Lemma plus_create : 
  Operations.is_binary_domain p x  ->  struct_property -> dplus create = p.
Proof. ir. apply Operations.binary_extensionality with x.
   pose (is_bin_plus_create H0). destruct i. am. am.
   app plus_create_when.
Qed.

Lemma mult_create: 
  Operations.is_binary_domain m x ->  struct_property -> dmult create = m.
Proof. ir. apply Operations.binary_extensionality with x.
   pose (is_bin_mult_create H0). destruct i.  am. am.
   app mult_create_when.
Qed.

Lemma neg_create: 
  Operations.is_unary_domain o x -> struct_property -> neg create = o.
Proof. intros. apply Operations.unary_extensionality with x.
  pose (is_un_neg_create H0). destruct i.  am. am. 
  intros. app neg_create_when.
Qed.

End Dconstruction.

Hint Rewrite underlying_create neg_create plus_create zero_create: mw.
Hint Rewrite one_create neg_create_when plus_create_when: mw.
Hint Rewrite mult_create mult_create_when: mw.

Lemma create_properties: forall x p o z u m,
    Operations.is_binary_domain p x ->
    Operations.is_unary_domain o x -> Operations.is_binary_domain m x ->
    struct_property x p o z u m ->
    ((dplus (create x p o z u m) = p) &
    (dmult (create x p o z u m) = m) &
    (neg (create x p o z u m) = o)).
Proof. ir. ee; mw.  
Qed.

Definition like (a:Bset):= 
  a = create (Ul a) (dplus a) (neg a) (zero a)  (one a)  (dmult a).

Lemma create_like: forall a, 
  like(create (Ul a) (dplus a) (neg a) (zero a)  (one a) (dmult a)).
Proof. ir. uf like.
  set (x:= create (Ul a) (dplus a) (neg a) (zero a)  (one a) (dmult a)).
  assert (Ul x = Ul a). uf x; mw.
  assert (zero x = zero a). uf x; mw.
  assert (one x = one a). uf x; mw.
  assert (neg x = neg a). uf x.
  apply Operations.unary_extensionality with (Ul x).  rw H.
  app is_un_domain_neg. app is_un_domain_neg. sy; am. ir. rwi H H2. mw. 
  assert (dplus x = dplus a). unfold x. 
  apply Operations.binary_extensionality with (Ul x). rw H.
  app is_bin_domain_plus. app is_bin_domain_plus. sy; am.
  ir. rewrite H in H3,H4. mw.
  assert (dmult x = dmult a). unfold x. 
  apply Operations.binary_extensionality with (Ul x). rw H.
  app is_bin_domain_mult. app is_bin_domain_mult. sy; am.
  ir. rewrite H in H4,H5. mw.
  rw H; rw H0; rw H1; rw H2; rw H3; rw H4. tv.
Qed.

Lemma like_create: forall x p o z u m,
  Operations.is_binary_domain p x ->
  Operations.is_unary_domain o x -> Operations.is_binary_domain m x ->
  struct_property x p o z u m ->
  like (create x p o z u m).
Proof. ir. pose (create_properties H H0 H1 H2). ee.
  red. rw H3; rw H4; rw H5. mw.  
Qed.

End Rnotation.

Export Rnotation.

Module Fnotation. (* Sfield notations *)

Section Fconstruction.
  Variables (x:Bset) (p:EEE)(o: EE) (z:Bset)(u:Bset) (m:EEE)(i:EE).

Definition create :=
  (denote Inv_not (unary x i) (Rnotation.create x p o z u m)).

Lemma underlying_create: Ul create = x.
Proof. ir. transitivity (Ul (Rnotation.create x p o z u m)). uf Ul. uf create. 
  drwe. tv. mw.
Qed.

Lemma neg_create_when: forall a, inc a x -> neg create a = o a.
Proof. ir. transitivity (neg  (Rnotation.create x p o z u m) a).
  uf create; uf neg; uf Ul. drwe. tv. mw.  
Qed.

Lemma plus_create_when: forall a b, inc a x -> inc b x -> 
  dplus create a b = (p a b).
Proof. ir. transitivity (dplus (Rnotation.create x p o z u m) a b).
  uf create; uf dplus; uf Ul. drwe. tv. mw. 
Qed.

Lemma zero_create: zero create  = z.
Proof. ir. transitivity (zero (Rnotation.create x p o z u m)). 
  uf zero. uf create. drwe. tv. mw.
Qed.

Lemma mult_create_when: forall a b, inc a x -> inc b x -> 
  dmult create a b = (m a b).
Proof. ir. transitivity (dmult (Rnotation.create x p o z u m) a b).
  uf create; uf dmult; uf Ul. drwe. tv. mw. 
Qed.

Lemma one_create: one create = u.
Proof. ir. transitivity (one (Rnotation.create x p o z u m)). 
  uf one. uf create. drwe. tv. mw.
Qed.

Lemma inv_create_when: forall a, inc a x -> inv create a = i a.
Proof. ir. pose (underlying_create). ufi create e.
  uf inv; uf create. rw e. drwe. app Operations.unary_recovers.
Qed.

Definition struct_property:=
  (Operations.is_unary_function i x) &
  (Rnotation.struct_property x p o z u m).

Lemma inc_create_zero_U: struct_property -> inc (zero create) (Ul create).
Proof. ir. rw underlying_create. rw zero_create. 
  destruct H. destruct H0. ee. destruct H2. am.
Qed.

Lemma is_bin_plus_create : struct_property
  -> Operations.is_strong_binary_function (dplus create) x.
Proof. ir. destruct H as [_ [_ [_ H]]]. 
  destruct H as [_ [_ H]]. red in H. red. split. 
   red. intros. rww plus_create_when. app H.  
   pose (H1:=underlying_create). app is_bin_domain_plus.
Qed.

Lemma is_bin_mult_create : struct_property
  -> Operations.is_strong_binary_function (dmult create) x.
Proof. ir. destruct H as [_ [_ [H _]]]. red in H. red. split. 
   red. intros. rww mult_create_when. app H.  
   pose (H1:=underlying_create). app is_bin_domain_mult.
Qed.

Lemma is_un_neg_create : struct_property
  -> Operations.is_strong_unary_function (neg create) x.
Proof. intros. destruct H as [_ [_ [_ H]]]. destruct H as [_ [H _]]. 
  red in H. red. split. 
  red. intros. rww neg_create_when. app H. 
  pose (H1:=underlying_create). app is_un_domain_neg.
Qed.

Lemma is_un_inv_create : struct_property
  -> Operations.is_strong_unary_function (inv create) x.
Proof. intros. destruct H. red in H. red. split. 
  red. intros. rww inv_create_when. app H. 
  pose (H1:=underlying_create). app is_un_domain_inv.
Qed.

Lemma plus_create : 
  Operations.is_binary_domain p x  ->  struct_property -> dplus create = p.
Proof. ir. apply Operations.binary_extensionality with x.
   pose (is_bin_plus_create H0). destruct i0.  am. am.
   app plus_create_when.
Qed.

Lemma mult_create : 
  Operations.is_binary_domain m x ->  struct_property -> dmult create = m.
Proof. ir. apply Operations.binary_extensionality with x.
   pose (is_bin_mult_create H0). destruct i0.  am. am.
   app mult_create_when.
Qed.

Lemma neg_create : 
  Operations.is_unary_domain o x -> struct_property -> neg create = o.
Proof. intros. apply Operations.unary_extensionality with x.
  pose (is_un_neg_create H0). destruct i0.  am. am. 
  intros. app neg_create_when.
Qed.


Lemma inv_create : 
  Operations.is_unary_domain i x -> struct_property -> inv create = i.
Proof. intros. apply Operations.unary_extensionality with x.
  pose (is_un_inv_create H0). destruct i0.  am. am. 
  intros. app inv_create_when.
Qed.

End Fconstruction.


Hint Rewrite underlying_create neg_create plus_create zero_create: mw.
Hint Rewrite one_create neg_create_when plus_create_when: mw.
Hint Rewrite mult_create mult_create_when: mw.
Hint Rewrite inv_create inv_create_when: mw.

Lemma create_properties:
  forall (x:Bset) (p:EEE)(o: EE)(z:Bset)(u:Bset) (m:EEE)(i:EE),
    Operations.is_unary_domain i x -> Operations.is_binary_domain p x ->
    Operations.is_unary_domain o x -> Operations.is_binary_domain m x ->
    struct_property x p o z u m i ->
    ((dplus (create x p o z u m i) = p) &
    (dmult (create x p o z u m i) = m) &
    (neg (create x p o z u m i) = o) &
    (inv (create x p o z u m i) = i)).
Proof. ir. ee;mw. 
Qed.

Definition like (a:Bset):= a = create (Ul a) (dplus a) (neg a) (zero a) 
  (one a) (dmult a) (inv a).

Lemma create_like: forall a, 
  like(create (Ul a) (dplus a) (neg a) (zero a)  (one a) (dmult a) (inv a)).
Proof. ir. uf like.
  set (x:=create (Ul a) (dplus a) (neg a) (zero a)  (one a) (dmult a) (inv a)).
  assert (Ul x = Ul a). uf x; mw.
  assert (zero x = zero a). uf x; mw.
  assert (one x = one a). uf x; mw.
  assert (inv x = inv a). uf x.
  apply Operations.unary_extensionality with (Ul x).  rw H.
  app is_un_domain_inv. app is_un_domain_inv. sy; am. ir. rwi H H2. mw.
  assert (neg x = neg a). uf x.
  apply Operations.unary_extensionality with (Ul x).  rw H.
  app is_un_domain_neg. app is_un_domain_neg. sy; am. ir. rwi H H3. mw.
  assert (dplus x = dplus a). unfold x. 
  apply Operations.binary_extensionality with (Ul x).  rw H.
  app is_bin_domain_plus. app is_bin_domain_plus. symmetry; am.
  ir. rewrite H in H4,H5. mw.
  assert (dmult x = dmult a). unfold x. 
  apply Operations.binary_extensionality with (Ul x).  rw H.
  app is_bin_domain_mult. app is_bin_domain_mult. symmetry; am.
  ir. rewrite H in H5,H6. mw.
  rw H; rw H0; rw H1; rw H2; rw H3; rw H4; rw H5. tv.
Qed.

Lemma like_create:  forall x p o z u m i,
    Operations.is_unary_domain i x -> Operations.is_binary_domain p x ->
    Operations.is_unary_domain o x -> Operations.is_binary_domain m x ->
    struct_property x p o z u m i ->
    like (create x p o z u m i).
Proof. ir. pose (create_properties H H0 H1 H2 H3). ee.
  red. rw H4; rw H5; rw H6; rw H7.  mw.
Qed.


End Fnotation.
 
Export Fnotation.

Module LMnotation.
Section Construction.
Variables (x:Bset)(y:Bset)(p:EEE)(o:EE)(z:Bset)(m:EEE).

Definition create :=
  denote Laux_not y
  (denote Laction_not (Operations.actcreate (Ul y) x m) 
    (Gnotation.create x p o z)).

Lemma underlying_create: Ul create = x.
Proof. ir. transitivity (Ul (Gnotation.create x p o z)). 
  uf Ul; uf create. drwe. tv. mw.
Qed.


Lemma zero_create: zero create  = z.
Proof. ir. transitivity (zero (Gnotation.create x p o z)). 
  uf zero; uf create. drw. mw.
Qed.

Lemma neg_create_when: forall a, inc a x -> neg create a = o a.
Proof. ir. transitivity (neg (Gnotation.create x p o z) a).
  uf create; uf neg; uf Ul. drwe.  tv.  mw.  
Qed.

Lemma plus_create_when: forall a b, inc a x -> inc b x -> 
  dplus create a b = (p a b).
Proof. ir. transitivity (dplus (Gnotation.create x p o z) a b).
  uf create; uf dplus; uf Ul. drwe. tv. mw.
Qed.

Lemma auxiliary_create: lauxiliary create = y.
Proof. ir. uf lauxiliary. uf create. drw.
Qed.

Lemma lmult_create_when: forall a b, inc a (Ul y) -> inc b x -> 
  lmult create a b = (m a b).
Proof. ir. uf lmult. rw underlying_create. rw auxiliary_create. uf create. 
  drw. app Operations.action_recovers. 
Qed.

Definition struct_property:=
  (inc z x) &
  (Operations.is_unary_function o x) &
  (Operations.is_binary_function p x) &
  (Operations.is_action_function m (Ul y) x).

Lemma inc_create_zero_U: struct_property -> inc (zero create) (Ul create).
Proof. ir. rw zero_create. rw underlying_create. destruct H. am.
Qed.

Lemma is_bin_plus_create : struct_property
  -> Operations.is_strong_binary_function (dplus create) x.
Proof. ir. red in H. ee. red. split. 
  red. intros. rww plus_create_when. app H1.  
  pose (underlying_create). app is_bin_domain_plus.
Qed.

Lemma is_act_mult_create : struct_property
  -> Operations.is_strong_action_function (lmult create) (Ul y) x.
Proof. ir. red in H. ee. red. split. 
   red. intros. rww lmult_create_when. app H2.  
   pose (underlying_create). app is_act_domain_lmult. app auxiliary_create.
Qed.

Lemma is_un_neg_create : struct_property
  -> Operations.is_strong_unary_function (neg create) x.
Proof. ir. red in H. ee. red. split.
  red. intros. rww neg_create_when. app H0. 
  pose (underlying_create). app is_un_domain_neg.
Qed.

Lemma plus_create : 
  Operations.is_binary_domain p x  ->  struct_property -> dplus create = p.
Proof. ir. apply Operations.binary_extensionality with x.
   pose (is_bin_plus_create H0). destruct i. am. am.
   app plus_create_when.
Qed.

Lemma lmult_create : 
  Operations.is_action_domain m (Ul y) x -> struct_property -> lmult create = m.
Proof. ir. apply Operations.action_extensionality with (Ul y) x.
   pose (is_act_mult_create H0). destruct i. am. am.
   app lmult_create_when.
Qed.

Lemma neg_create: 
  Operations.is_unary_domain o x -> struct_property -> neg create = o.
Proof. intros. apply Operations.unary_extensionality with x.
  pose (is_un_neg_create H0). destruct i.  am. am. 
  intros. app neg_create_when.
Qed.

End Construction.

Hint Rewrite underlying_create auxiliary_create 
  neg_create plus_create zero_create: mw.
Hint Rewrite neg_create_when plus_create_when: mw.
Hint Rewrite lmult_create lmult_create_when: mw.



Lemma create_properties:
  forall (x:Bset) (y:Bset)(p:EEE)(o: EE)(z:Bset)(m:EEE),
    Operations.is_binary_domain p x ->
    Operations.is_unary_domain o x -> 
    Operations.is_action_domain m (Ul y) x ->
    struct_property x y p o z m ->
    ((dplus (create x y p o z m) = p) &
    (lmult (create x y p o z m) = m) &
    (neg (create x y p o z m) = o)).
Proof. ir. ee; mw.
Qed.

Definition like (a:Bset):= a = create (Ul a) (lauxiliary a) (dplus a) (neg a) 
  (zero a) (lmult a).

Lemma create_like: forall a, 
  like(create (Ul a) (lauxiliary a) (dplus a) (neg a) (zero a) (lmult a)).
Proof. ir. uf like.
  set (x:= create (Ul a) (lauxiliary a) (dplus a) (neg a) (zero a) (lmult a)).
  assert (Ul x = Ul a). uf x; mw.
  assert (zero x = zero a). uf x; mw.
  assert (neg x = neg a). uf x.
  apply Operations.unary_extensionality with (Ul x).  rw H.
  app is_un_domain_neg. app is_un_domain_neg. sy; am. ir. rwi H H1. mw. 
  assert (dplus x = dplus a). unfold x. 
  apply Operations.binary_extensionality with (Ul x).  rw H.
  app is_bin_domain_plus. app is_bin_domain_plus. symmetry; am.
  ir. rewrite H in H2,H3. mw. 
  assert (lauxiliary x = lauxiliary a). uf x; mw.
  assert (lmult x = lmult a). unfold x. 
  set (y:= lauxiliary x).
  apply Operations.action_extensionality with (Ul y) (Ul x).  rw H.
  app is_act_domain_lmult. app is_act_domain_lmult. sy; am. wrr H3.
  ir. rewrite H in H5.  unfold y in H4. rewrite H3 in H4. mw. 
  rw H; rw H0; rw H1; rw H2; rw H3; rw H4. tv.
Qed.

Lemma like_create:  forall x y p o z m,
  Operations.is_binary_domain p x ->
  Operations.is_unary_domain o x -> Operations.is_action_domain m (Ul y) x ->
  struct_property x y p o z m ->
  like (create x y p o z m).
Proof. ir. pose (create_properties H H0 H1 H2). ee.
  red. rw H3; rw H4; rw H5. mw. 
Qed.

End LMnotation.
 
Export LMnotation.

Module RMnotation.
Section Construction.
Variables (x:Bset)(y:Bset)(p:EEE)(o:EE)(z:Bset)(m:EEE).

Definition create :=
  denote Raux_not y
  (denote Raction_not (Operations.actcreate (Ul y) x m) 
    (Gnotation.create x p o z)).

Lemma underlying_create: Ul create = x.
Proof. ir. transitivity (Ul (Gnotation.create x p o z)). 
  uf Ul; uf create. drwe. tv. mw.
Qed.

Lemma zero_create: zero create  = z.
Proof. ir. transitivity (zero (Gnotation.create x p o z)). 
  uf zero; uf create. drw. mw.
Qed.

Lemma neg_create_when: forall a, inc a x -> neg create a = o a.
Proof. ir. transitivity (neg (Gnotation.create x p o z) a).
  uf create; uf neg; uf Ul. drwe.  tv.  mw.  
Qed.

Lemma plus_create_when: forall a b, inc a x -> inc b x -> 
  dplus create a b = (p a b).
Proof. ir. transitivity (dplus (Gnotation.create x p o z) a b).
  uf create; uf dplus; uf Ul. drwe. tv. mw.
Qed.

Lemma auxiliary_create: rauxiliary create = y.
Proof. ir. uf rauxiliary. uf create. drw.
Qed.

Lemma rmult_create_when: forall a b, inc a (Ul y) -> inc b x -> 
  rmult create a b = (m a b).
Proof. ir. uf rmult. rw underlying_create. rw auxiliary_create. uf create. 
  drw. app Operations.action_recovers. 
Qed.

Definition struct_property:=
  (inc z x) &
  (Operations.is_unary_function o x) &
  (Operations.is_binary_function p x) &
  (Operations.is_action_function m (Ul y) x).


Lemma inc_create_zero_U: struct_property -> inc (zero create) (Ul create).
Proof. ir. rw zero_create. rw underlying_create. destruct H. am.
Qed.

Lemma is_bin_plus_create : struct_property
  -> Operations.is_strong_binary_function (dplus create) x.
Proof. ir. red in H. ee. red. split. 
  red. intros. rww plus_create_when. app H1.  
  pose (underlying_create). app is_bin_domain_plus.
Qed.

Lemma is_act_mult_create : struct_property
  -> Operations.is_strong_action_function (rmult create) (Ul y) x.
Proof. ir. red in H. ee. red. split. 
   red. intros. rww rmult_create_when. app H2.  
   pose (underlying_create). app is_act_domain_rmult. app auxiliary_create.
Qed.

Lemma is_un_neg_create : struct_property
  -> Operations.is_strong_unary_function (neg create) x.
Proof. ir. red in H. ee. red. split.
  red. intros. rww neg_create_when. app H0. 
  pose (underlying_create). app is_un_domain_neg.
Qed.

Lemma plus_create : 
  Operations.is_binary_domain p x  ->  struct_property -> dplus create = p.
Proof. ir. apply Operations.binary_extensionality with x.
   pose (is_bin_plus_create H0). destruct i. am. am.
   app plus_create_when.
Qed.

Lemma rmult_create : 
  Operations.is_action_domain m (Ul y) x -> struct_property -> rmult create = m.
Proof. ir. apply Operations.action_extensionality with (Ul y) x.
   pose (is_act_mult_create H0). destruct i. am. am.
   app rmult_create_when.
Qed.

Lemma neg_create: 
  Operations.is_unary_domain o x -> struct_property -> neg create = o.
Proof. intros. apply Operations.unary_extensionality with x.
  pose (is_un_neg_create H0). destruct i.  am. am. 
  intros. app neg_create_when.
Qed.

End Construction.

Hint Rewrite underlying_create auxiliary_create 
  neg_create plus_create zero_create: mw.
Hint Rewrite neg_create_when plus_create_when: mw.
Hint Rewrite rmult_create rmult_create_when: mw.



Lemma create_properties:
  forall (x:Bset) (y:Bset)(p:EEE)(o: EE)(z:Bset)(m:EEE),
    Operations.is_binary_domain p x ->
    Operations.is_unary_domain o x -> 
    Operations.is_action_domain m (Ul y) x ->
    struct_property x y p o z m ->
    ((dplus (create x y p o z m) = p) &
    (rmult (create x y p o z m) = m) &
    (neg (create x y p o z m) = o)).
Proof. ir. ee; mw.
Qed.

Definition like (a:Bset):= a = create (Ul a) (rauxiliary a) (dplus a) (neg a) 
  (zero a) (rmult a).

Lemma create_like: forall a, 
  like(create (Ul a) (rauxiliary a) (dplus a) (neg a) (zero a) (rmult a)).
Proof. ir. uf like.
  set (x:= create (Ul a) (rauxiliary a) (dplus a) (neg a) (zero a) (rmult a)).
  assert (Ul x = Ul a). uf x; mw.
  assert (zero x = zero a). uf x; mw.
  assert (neg x = neg a). uf x.
  apply Operations.unary_extensionality with (Ul x).  rw H.
  app is_un_domain_neg. app is_un_domain_neg. sy; am. ir. rwi H H1. mw. 
  assert (dplus x = dplus a). unfold x. 
  apply Operations.binary_extensionality with (Ul x).  rw H.
  app is_bin_domain_plus. app is_bin_domain_plus. symmetry; am.
  ir. rewrite H in H2,H3. mw. 
  assert (rauxiliary x = rauxiliary a). uf x; mw.
  assert (rmult x = rmult a). unfold x. 
  set (y:= rauxiliary x).
  apply Operations.action_extensionality with (Ul y) (Ul x).  rw H.
  app is_act_domain_rmult. app is_act_domain_rmult. sy; am. wrr H3.
  ir. rewrite H in H5.  unfold y in H4. rewrite H3 in H4. mw. 
  rw H; rw H0; rw H1; rw H2; rw H3; rw H4. tv.
Qed.

Lemma like_create:  forall x y p o z m,
  Operations.is_binary_domain p x ->
  Operations.is_unary_domain o x -> Operations.is_action_domain m (Ul y) x ->
  struct_property x y p o z m ->
  like (create x y p o z m).
Proof. ir. pose (create_properties H H0 H1 H2). ee.
  red. rw H3; rw H4; rw H5. mw. 
Qed.

End RMnotation.
 
Export RMnotation.

Module LRMnotation.
Section Construction.
Variables (x:Bset)(yl yr:Bset)(p:EEE)(o:EE)(z:Bset)(ml mr:EEE).

Definition create :=
  denote Laux_not yl
    (denote Raux_not yr
      (denote Laction_not (Operations.actcreate (Ul yl) x ml) 
        (denote Raction_not (Operations.actcreate (Ul yr) x mr) 
          (Gnotation.create x p o z)))).

Lemma lauxiliary_create: lauxiliary create = yl.
Proof. ir. uf lauxiliary. uf create. drw.
Qed.

Lemma rauxiliary_create: rauxiliary create = yr.
Proof. ir. uf rauxiliary. uf create. drw.
Qed.


Lemma underlying_create: Ul create = x.
Proof. ir. transitivity (Ul (Gnotation.create x p o z)). 
  uf Ul; uf create. drwe. tv. mw.
Qed.


Lemma zero_create: zero create  = z.
Proof. ir. transitivity (zero (Gnotation.create x p o z)). 
  uf zero; uf create. drw. mw.
Qed.

Lemma neg_create_when: forall a, inc a x -> neg create a = o a.
Proof. ir. transitivity (neg (Gnotation.create x p o z) a).
  uf create; uf neg; uf Ul. drwe.  tv.  mw.  
Qed.

Lemma plus_create_when: forall a b, inc a x -> inc b x -> 
  dplus create a b = (p a b).
Proof. ir. transitivity (dplus (Gnotation.create x p o z) a b).
  uf create; uf dplus; uf Ul. drwe. tv. mw.
Qed.

Lemma lmult_create_when: forall a b, inc a (Ul yl) -> inc b x -> 
  lmult create a b = (ml a b).
Proof. ir. uf lmult. rw underlying_create. rw lauxiliary_create. uf create. 
  drw. app Operations.action_recovers. 
Qed.

Lemma rmult_create_when: forall a b, inc a (Ul yr) -> inc b x -> 
  rmult create a b = (mr a b).
Proof. ir. uf rmult. rw underlying_create. rw rauxiliary_create. uf create. 
  drw. app Operations.action_recovers. 
Qed.

Definition struct_property:=
  (inc z x) &
  (Operations.is_unary_function o x) &
  (Operations.is_binary_function p x) &
  (Operations.is_action_function ml (Ul yl) x) &
  (Operations.is_action_function mr (Ul yr) x).

Lemma inc_create_zero_U: struct_property -> inc (zero create) (Ul create).
Proof. ir. rewrite zero_create. rewrite underlying_create. destruct H. am.
Qed.

Lemma is_bin_plus_create : struct_property
  -> Operations.is_strong_binary_function (dplus create) x.
Proof. ir. red in H. ee. red. split. 
   red. intros. rww plus_create_when. app H1.  
   pose (underlying_create). app is_bin_domain_plus.
Qed.

Lemma is_act_lmult_create : struct_property
  -> Operations.is_strong_action_function (lmult create) (Ul yl) x.
Proof. ir. red in H. ee. red. split. 
   red. intros. rww lmult_create_when. app H2.  
   pose (underlying_create). app is_act_domain_lmult. app lauxiliary_create.
Qed.

Lemma is_act_rmult_create : struct_property
  -> Operations.is_strong_action_function (rmult create) (Ul yr) x.
Proof. ir. red in H. ee. red. split. 
   red. intros. rww rmult_create_when. app H3.  
   pose (underlying_create). app is_act_domain_rmult. app rauxiliary_create.
Qed.

Lemma is_un_neg_create : struct_property
  -> Operations.is_strong_unary_function (neg create) x.
Proof. ir. red in H. ee. red. split.
  red. intros. rww neg_create_when. app H0. 
  pose (underlying_create). app is_un_domain_neg.
Qed.

Lemma plus_create : 
  Operations.is_binary_domain p x  ->  struct_property -> dplus create = p.
Proof. ir. apply Operations.binary_extensionality with x.
   pose (is_bin_plus_create H0). destruct i. am. am.
   app plus_create_when.
Qed.

Lemma lmult_create : Operations.is_action_domain ml (Ul yl) x -> 
  struct_property -> lmult create = ml.
Proof. ir. apply Operations.action_extensionality with (Ul yl) x.
   pose (is_act_lmult_create H0). destruct i. am. am.
   app lmult_create_when.
Qed.


Lemma rmult_create :  Operations.is_action_domain mr (Ul yr) x -> 
  struct_property -> rmult create = mr.
Proof. ir. apply Operations.action_extensionality with (Ul yr) x.
   pose (is_act_rmult_create H0). destruct i. am. am.
   app rmult_create_when.
Qed.



Lemma neg_create: 
  Operations.is_unary_domain o x -> struct_property -> neg create = o.
Proof. intros. apply Operations.unary_extensionality with x.
  pose (is_un_neg_create H0). destruct i.  am. am. 
  intros. app neg_create_when.
Qed.

End Construction.

Hint Rewrite underlying_create lauxiliary_create rauxiliary_create 
  neg_create plus_create zero_create: mw.
Hint Rewrite one_create neg_create_when plus_create_when: mw.
Hint Rewrite lmult_create lmult_create_when: mw.
Hint Rewrite rmult_create rmult_create_when: mw.



Lemma create_properties:
  forall (x:Bset) (yl yr:Bset)(p:EEE)(o: EE)(z:Bset)(ml mr:EEE),
    Operations.is_binary_domain p x ->
    Operations.is_unary_domain o x -> 
    Operations.is_action_domain ml (Ul yl) x ->
    Operations.is_action_domain mr (Ul yr) x ->
    struct_property x yl yr p o z ml mr ->
    ((dplus (create x yl yr p o z ml mr) = p) &
    (lmult (create x yl yr p o z ml mr) = ml) &
    (rmult (create x yl yr p o z ml mr) = mr) &
    (neg (create x yl yr p o z ml mr) = o)).
Proof. ir. ee; mw.
Qed.

Definition like (a:Bset):= a = create (Ul a) (lauxiliary a) (rauxiliary a) 
  (dplus a) (neg a) (zero a) (lmult a) (rmult a).

Lemma create_like: forall a, 
  like(create (Ul a) (lauxiliary a) (rauxiliary a) 
  (dplus a) (neg a) (zero a) (lmult a) (rmult a)).
Proof. ir. uf like.
  set (x:= create (Ul a) (lauxiliary a) (rauxiliary a) 
  (dplus a) (neg a) (zero a) (lmult a) (rmult a)).
  assert (Ul x = Ul a). uf x; mw.
  assert (zero x = zero a). uf x; mw.
  assert (neg x = neg a). uf x.
  apply Operations.unary_extensionality with (Ul x).  rw H.
  app is_un_domain_neg. app is_un_domain_neg. sy; am. ir. rwi H H1. mw. 
  assert (dplus x = dplus a). unfold x. 
  apply Operations.binary_extensionality with (Ul x).  rw H.
  app is_bin_domain_plus. app is_bin_domain_plus. symmetry; am.
  ir. rewrite H in H2,H3. mw. 
  assert (lauxiliary x = lauxiliary a). uf x; mw.
  assert (rauxiliary x = rauxiliary a). uf x; mw.
  assert (lmult x = lmult a). unfold x. 
  set (y:= lauxiliary x).
  apply Operations.action_extensionality with (Ul y) (Ul x).  rw H.
  app is_act_domain_lmult. app is_act_domain_lmult. sy; am. wrr H3.
  ir. rewrite H in H6.  unfold y in H5. rewrite H3 in H5. mw. 
  assert (rmult x = rmult a). unfold x. 
  set (y:= rauxiliary x).
  apply Operations.action_extensionality with (Ul y) (Ul x).  rw H.
  app is_act_domain_rmult. app is_act_domain_rmult. sy; am. wrr H4.
  ir. rewrite H in H7.  unfold y in H6. rewrite H4 in H6. mw. 
  rw H; rw H0; rw H1; rw H2; rw H3; rw H4; rw H5; rw H6. tv.
Qed.

Lemma like_create:  forall x yl yr p o z ml mr,
  Operations.is_binary_domain p x ->
  Operations.is_unary_domain o x -> Operations.is_action_domain ml (Ul yl) x ->
  Operations.is_action_domain mr (Ul yr) x ->
  struct_property x yl yr p o z ml mr ->
  like (create x yl yr p o z ml mr).
Proof. ir. pose (create_properties H H0 H1 H2 H3). ee.
  red. rw H4; rw H5; rw H6; rw H7. mw. 
Qed.

End LRMnotation.
 
Export LRMnotation.

(* ******************************************************************* *)
Module Definitions.


Section Construction.
Variables (x:Bset) (p: EEE)(o: EE)(z:Bset)(u:Bset)(m:EEE)(i:EE).

Definition plus_commutativity  :=
  forall (a b:Bset), inc a x -> inc b x -> p a b = p b a.
Definition plus_associativity  :=
  forall (a b c:Bset), inc a x -> inc b x -> inc c x-> 
    p (p a b) c = p a (p b c).
Definition mult_commutativity  :=
  forall (a b:Bset), inc a x -> inc b x -> m a b = m b a.
Definition mult_associativity  :=
  forall (a b c:Bset), inc a x -> inc b x -> inc c x-> 
    m (m a b) c = m a (m b c).
Definition plus_unit_left :=
 forall (a:Bset), inc a x -> p z a =a. 
Definition plus_unit_right :=
 forall (a:Bset), inc a x -> p a z =a. 
Definition mult_unit_left :=
 forall (a:Bset), inc a x -> m u a =a. 
Definition mult_unit_right :=
 forall (a:Bset), inc a x -> m a u =a. 
Definition plus_inverse_left  := 
  forall (a:Bset), inc a x -> p (o a) a= z. 
Definition plus_inverse_right  := 
  forall (a:Bset), inc a x -> p a (o a) = z. 
Definition mult_inverse_left  := 
  forall (a:Bset), inc a x -> a <> z -> m (i a) a= u. 
Definition mult_inverse_right  := 
  forall (a:Bset), inc a x -> a <> z -> m a (i a) = u. 
Definition left_distributive :=
  forall(a b c:Bset), inc a x -> inc b x -> inc c x ->
    m a (p b c) = p (m a b) (m a c).
Definition right_distributive :=
  forall(a b c:Bset), inc a x -> inc b x -> inc c x ->
    m (p b c) a = p (m b a) (m c a).

Definition plus_internal:= Operations.is_binary_function p x.
Definition neg_internal:= Operations.is_unary_function o x.
Definition mult_internal:= Operations.is_binary_function m x.
Definition inv_internal:= Operations.is_unary_function i x.


Definition mproperty  :=
  (inc z x)  &
  plus_internal &
  plus_associativity &
  plus_unit_right & plus_unit_left.

Definition gproperty  :=
  mproperty &
  neg_internal &
  plus_inverse_right & plus_inverse_left.

Definition cgproperty  :=
  gproperty & plus_commutativity.

Definition rproperty := 
  cgproperty &
  mult_internal & (inc u x) & 
  mult_associativity &
  mult_unit_right & mult_unit_left &
  left_distributive & right_distributive.

Definition fproperty :=
  rproperty & mult_inverse_left & mult_inverse_right & inv_internal
  &(exists a, inc a x & a <> z).

Definition crproperty := rproperty & mult_commutativity.
Definition cfproperty := fproperty & mult_commutativity.

End Construction.
Section Mconstruction.
Variables (x:Bset) (y:Bset) (p: EEE)(o: EE)(z:Bset)(m:EEE).

Definition action_left_distributive :=
  forall(a b c:Bset), inc a (Ul y) -> inc b x -> inc c x ->
    m a (p b c) = p (m a b) (m a c).
Definition action_right_distributive :=
  forall(a b c:Bset), inc a x -> inc b (Ul y) -> inc c (Ul y) ->
    m (dplus y  b c) a = p (m b a) (m c a).
Definition laction_associative :=
  forall(a b c:Bset), inc a (Ul y) -> inc b (Ul y) -> inc c x ->
    m a (m b c)  = m (dmult y a b) c.

Definition raction_associative :=
  forall(a b c:Bset), inc a (Ul y) -> inc b (Ul y) -> inc c x ->
    m a (m b c)  = m (dmult y b a) c.

Definition action_unary:=
  forall a:Bset, inc a x -> m (one y) a = a.
  
Definition action_internal:= Operations.is_action_function m (Ul y) x.
End Mconstruction.

Definition lmproperty (x:Bset) (y:Bset) (p: EEE)(o: EE)(z:Bset)(m:EEE):=
  cgproperty x p o z & action_left_distributive x y p m & 
  action_right_distributive x y p m&
  laction_associative x y m & action_internal x y m & action_unary x y m.

Definition rmproperty (x:Bset) (y:Bset) (p: EEE)(o: EE)(z:Bset)(m:EEE):=
  cgproperty x p o z & action_left_distributive x y p m & 
  action_right_distributive x y p m&
  raction_associative x y m & action_internal x y m & action_unary x y m.

Definition lrmproperty (x:Bset) (yl yr:Bset) (p: EEE)(o: EE)(z:Bset)(ml mr:EEE):=
  cgproperty x p o z & 
  action_left_distributive x yl p ml & action_right_distributive x yl p ml &
  action_left_distributive x yr p mr & action_right_distributive x yr p mr&
  laction_associative x yl ml &  raction_associative x yr mr & 
  action_internal x yl ml & action_unary x yl ml &
  action_internal x yr mr & action_unary x yr mr.

Definition monoid_axioms a := 
   mproperty  (Ul a) (dplus a) (zero a).

Definition strong_monoid_axioms a := 
  (Monoid_notation.like a) & (monoid_axioms a).

Definition group_axioms a := 
   gproperty  (Ul a) (dplus a) (neg a) (zero a).

Definition strong_group_axioms a := 
   (Gnotation.like a) & (group_axioms a).

Definition cgroup_axioms a := 
  cgproperty (Ul a) (dplus a) (neg a) (zero a).

Definition strong_cgroup_axioms a := 
  (Gnotation.like a) &(cgroup_axioms a).

Definition ring_axioms a := 
   rproperty  (Ul a) (dplus a) (neg a) (zero a) (one a) (dmult a).

Definition strong_ring_axioms a := 
  (Rnotation.like a) & (ring_axioms a).

Definition cring_axioms a := 
  crproperty (Ul a) (dplus a) (neg a) (zero a) (one a) (dmult a).

Definition strong_cring_axioms a := 
  (Rnotation.like a) &(cring_axioms a).

Definition sfield_axioms a := 
  fproperty  (Ul a) (dplus a) (neg a) (zero a) (one a) (dmult a) (inv a).

Definition strong_sfield_axioms a := 
  (Fnotation.like a) & (sfield_axioms a).

Definition field_axioms a := 
  cfproperty  (Ul a) (dplus a) (neg a) (zero a) (one a) (dmult a) (inv a).

Definition strong_field_axioms a := 
  (Fnotation.like a) & (field_axioms a).

Definition lmodule_axioms a :=
  (ring_axioms (lauxiliary a)) &
  (lmproperty (Ul a) (lauxiliary a) (dplus a) (neg a) (zero a) (lmult a)).

Definition strong_lmodule_axioms a :=
  (LMnotation.like a) & (Rnotation.like (lauxiliary a)) & (lmodule_axioms a).

Definition rmodule_axioms a :=
  (ring_axioms (rauxiliary a)) &
  (rmproperty (Ul a) (rauxiliary a) (dplus a) (neg a) (zero a) (rmult a)).

Definition strong_rmodule_axioms a :=
  (RMnotation.like a) & (Rnotation.like (rauxiliary a)) & (rmodule_axioms a).

Definition lrmodule_axioms a :=
  (ring_axioms (lauxiliary a)) &  (ring_axioms (rauxiliary a)) &
  (lrmproperty (Ul a) (lauxiliary a) (rauxiliary a) (dplus a) (neg a) (zero a) 
    (lmult a) (rmult a)).

Definition strong_lrmodule_axioms a :=
  (LRMnotation.like a) &  (Rnotation.like (lauxiliary a)) & 
  (Rnotation.like (rauxiliary a)) & (lrmodule_axioms a).

Lemma monoid_implies_mstruct:forall x p z,
  mproperty x p z -> (Monoid_notation.struct_property x p z).
Proof. ir. red in H. ee. red. split; am.
Qed.

Lemma group_implies_gstruct:forall x p o z,
 gproperty x p o z -> (Gnotation.struct_property x p o z).
Proof. ir. red in H. ee. cp (monoid_implies_mstruct H). red in H3; ee. 
  red. intuition.
Qed.

Lemma ring_implies_gstruct:forall x p o z u m,
 rproperty x p o z u m-> (Gnotation.struct_property x p o z).
Proof. ir. destruct H. destruct H. app group_implies_gstruct.
Qed.

Lemma ring_implies_rstruct: forall x  p o z u m,
  rproperty x p o z u m -> Rnotation.struct_property x p o z u m.
Proof. ir. red. split. red in H. ee; am. split. red in H. ee. am.
  app (ring_implies_gstruct H).
Qed.

Lemma sfield_implies_rstruct: forall x p o z u m i,
  fproperty x p o z u m i-> Rnotation.struct_property x p o z u m.
Proof. ir. destruct H.  app ring_implies_rstruct.
Qed.

Lemma sfield_implies_fstruct: forall x p o z u m i,
  fproperty x p o z u m i-> Fnotation.struct_property x p o z u m i.
Proof. ir. destruct H. split.  destruct H0 as [_ [_ [H0 _]]]. am.
  app ring_implies_rstruct.
Qed.

Lemma lmodule_implies_mstruct: forall x y p o z m,
  lmproperty x y p o z m -> LMnotation.struct_property  x y p o z m.
Proof. ir. assert (Gnotation.struct_property x p o z).
  destruct H. destruct H. app group_implies_gstruct. red in H0. ee.
  red in H. ee. red. intuition. 
Qed.

Lemma rmodule_implies_mstruct: forall x y p o z m,
  rmproperty x y p o z m -> RMnotation.struct_property  x y p o z m.
Proof. ir. assert (Gnotation.struct_property x p o z).
  destruct H. destruct H. app group_implies_gstruct. red in H0. ee.
  red in H. ee. red. intuition. 
Qed.

Lemma lrmodule_implies_mstruct:  forall x yl yr p o z ml mr,
  lrmproperty x yl yr  p o z ml mr -> 
  LRMnotation.struct_property  x yl yr p o z ml mr.
Proof. ir. assert (Gnotation.struct_property x p o z).
  destruct H. destruct H. app group_implies_gstruct. red in H0. ee.
  red in H. ee. red. intuition. 
Qed.


Lemma lrmodule_implies_lmstruct:  forall x yl yr p o z ml mr,
  lrmproperty x yl yr  p o z ml mr -> 
  LMnotation.struct_property  x yl p o z ml.
Proof. ir. cp (lrmodule_implies_mstruct H). red in H0. ee. red. intuition.
Qed.

Lemma lrmodule_implies_rmstruct: forall x yl yr p o z ml mr,
  lrmproperty x yl yr  p o z ml mr -> 
  RMnotation.struct_property x yr p o z mr.
Proof. ir. cp (lrmodule_implies_mstruct H). red in H0. ee. red. intuition.
Qed.

Lemma mcreate_axioms: forall x p z,
  Operations.is_binary_domain p x -> mproperty x p z -> 
  strong_monoid_axioms (Monoid_notation.create x p z).
Proof. ir. 
  assert (Monoid_notation.struct_property x p z). app monoid_implies_mstruct.
  split. app Monoid_notation.like_create. red. mw. 
Qed.

Lemma monoid_is_monoid: forall a, monoid_axioms a -> 
  strong_monoid_axioms (Monoid_notation.create  (Ul a) (dplus a) (zero a)).
Proof. intros. apply mcreate_axioms. app is_bin_domain_plus. am.
Qed.

Lemma gcreate_axioms: forall x p o z,
  Operations.is_binary_domain p x -> Operations.is_unary_domain o x ->
  gproperty x p o z -> 
  strong_group_axioms (Gnotation.create x p o z).
Proof. ir. cp (group_implies_gstruct H1). split.
  app Gnotation.like_create. red. mw.
Qed.

Lemma group_is_group: forall a, group_axioms a -> 
  strong_group_axioms (Gnotation.create  (Ul a) (dplus a) (neg a) (zero a)).
Proof. intros. apply gcreate_axioms. app is_bin_domain_plus.
  app is_un_domain_neg. am.
Qed.

Lemma group_is_monoid: forall a, group_axioms a -> monoid_axioms a.
Proof. ir. red in H. red in H. destruct H. am.
Qed.

Lemma strong_group_is_monoid: forall a, group_axioms a -> 
  strong_monoid_axioms (Monoid_notation.create (Ul a) (dplus a) (zero a)).
Proof. ir. destruct H.
 pose (b:=Monoid_notation.create (Ul a) (dplus a) (zero a)).
  app mcreate_axioms. app is_bin_domain_plus. 
Qed. 

Lemma cgcreate_axioms: forall x p o z,
  Operations.is_binary_domain p x -> Operations.is_unary_domain o x ->
  gproperty x p o z -> plus_commutativity x p ->
  strong_cgroup_axioms (Gnotation.create x p o z).
Proof. ir. pose (a:=Gnotation.create x p o z).
  assert (strong_group_axioms a). uf a. app gcreate_axioms.
  red in H3. ee. fold a. split. am. split. mw. 
  assert (Gnotation.struct_property x p o z). app group_implies_gstruct.
  unfold a. mw.
Qed.

Lemma plus_gcreate : forall a, group_axioms a-> 
  dplus (Gnotation.create  (Ul a) (dplus a) (neg a) (zero a)) = dplus a.
Proof. ir. red in H. ee. mw. app is_bin_domain_plus. app group_implies_gstruct.
Qed.

Lemma neg_gcreate : forall a, group_axioms a-> 
  neg (Gnotation.create  (Ul a) (dplus a) (neg a) (zero a)) = neg a.
Proof. ir. red in H. ee. mw. app is_un_domain_neg. app group_implies_gstruct.
Qed.

Lemma cgroup_is_group: forall a,  cgroup_axioms a -> group_axioms a.
Proof. intros. destruct H. red. am.
Qed.

Lemma strong_cgroup_is_group: forall a,  
  strong_cgroup_axioms a -> strong_group_axioms a.
Proof. intros. destruct H. pose (cgroup_is_group H0).  split; am.
Qed.

Lemma cgroup_is_cgroup: forall a, cgroup_axioms a ->
  strong_cgroup_axioms (Gnotation.create  (Ul a) (dplus a) (neg a) (zero a)).
Proof. intros. destruct H.
  app cgcreate_axioms.  app is_bin_domain_plus.  app is_un_domain_neg. 
Qed.

Lemma rcreate_axioms: forall x p o z u m,
  Operations.is_binary_domain p x -> Operations.is_unary_domain o x ->
  Operations.is_binary_domain m x ->
  rproperty x p o z u m-> 
  strong_ring_axioms (Rnotation.create x p o z u m).
Proof. ir. cp(ring_implies_rstruct H2).
  red. split.  app Rnotation.like_create. red. mw.
Qed.

Lemma ring_is_ring: forall a, ring_axioms a ->
  strong_ring_axioms (Rnotation.create (Ul a) (dplus a) (neg a) (zero a)
    (one a)  (dmult a)).
Proof. ir. ap rcreate_axioms. app is_bin_domain_plus.
  app is_un_domain_neg. app is_bin_domain_mult. am.
Qed.

Lemma ring_is_cgroup: forall a, ring_axioms a -> cgroup_axioms a.
Proof. ir. red in H. red in H. destruct H. am.
Qed.

Lemma strong_ring_is_cgroup: forall a, ring_axioms a -> 
  strong_cgroup_axioms (Gnotation.create (Ul a) (dplus a) (neg a) (zero a)).
Proof. ir. destruct H. destruct H.
 pose (b:=Gnotation.create (Ul a) (dplus a) (neg a) (zero a)).
  app cgcreate_axioms. app is_bin_domain_plus. app is_un_domain_neg. 
Qed. 

Lemma plus_rcreate : forall a, ring_axioms a-> 
  dplus (Rnotation.create  (Ul a) (dplus a) (neg a) (zero a)(one a)(dmult a)) 
  = dplus a.
Proof.  ir. mw. app is_bin_domain_plus. red in H. app ring_implies_rstruct.
Qed.

Lemma neg_rcreate : forall a, ring_axioms a-> 
  neg (Rnotation.create  (Ul a) (dplus a) (neg a) (zero a)(one a)(dmult a)) 
  = neg a.
Proof. ir. mw. app is_un_domain_neg.  red in H. app ring_implies_rstruct.
Qed.

Lemma mult_rcreate : forall a, ring_axioms a-> 
  dmult (Rnotation.create  (Ul a) (dplus a) (neg a) (zero a)(one a)(dmult a))
  = dmult a.
Proof. ir. mw. app is_bin_domain_mult. red in H. app ring_implies_rstruct.
Qed.

Lemma cring_is_ring: forall a,  cring_axioms a -> ring_axioms a.
Proof. intros. destruct H.  am. 
Qed.

Lemma strong_cring_is_ring: forall a, 
  strong_cring_axioms a -> strong_ring_axioms a.
Proof. intros. destruct H. pose (cring_is_ring H0). split;am. 
Qed.

Lemma cring_is_cring: forall a,  cring_axioms a ->
  strong_cring_axioms (Rnotation.create  (Ul a) (dplus a) (neg a) 
    (zero a)(one a)(dmult a)).
Proof. intros. destruct H.
  set (b:=Rnotation.create  (Ul a) (dplus a) (neg a) (zero a)(one a)(dmult a)).
  assert (strong_ring_axioms b). unfold b. app ring_is_ring. red. destruct H1.
  split. am. split. am.
  assert (Ul b = Ul a). unfold b. mw.
  assert (dmult b = dmult a). unfold b. app mult_rcreate.
  rw H3; rww H4.
Qed.

Lemma fcreate_axioms: forall x p o z u m i,
  Operations.is_binary_domain p x -> Operations.is_unary_domain o x ->
  Operations.is_binary_domain m x -> Operations.is_unary_domain i x ->
  fproperty x p o z u m i-> 
  strong_sfield_axioms (Fnotation.create x p o z u m i).
Proof. ir. cp (sfield_implies_fstruct H3).
  red. split. app Fnotation.like_create. red.  mw. 
Qed.

Lemma sfield_is_sfield: forall a, sfield_axioms a ->
  strong_sfield_axioms (Fnotation.create (Ul a) (dplus a) (neg a) (zero a) 
    (one a) (dmult a) (inv a)).
Proof. ir. ap fcreate_axioms. app is_bin_domain_plus.
  app is_un_domain_neg. app is_bin_domain_mult. app is_un_domain_inv. am.
Qed.

Lemma field_is_sfield: forall a,  field_axioms a -> sfield_axioms a.
Proof. intros. destruct H.  am. 
Qed.

Lemma mult_fcreate : forall a, sfield_axioms a-> 
  dmult (Fnotation.create  (Ul a) (dplus a) (neg a) (zero a)(one a)
    (dmult a)(inv a)) = dmult a.
Proof. ir. mw. app is_bin_domain_mult. red in H. app sfield_implies_fstruct.
Qed.

Lemma field_is_field: forall a,  field_axioms a ->
  strong_field_axioms (Fnotation.create  (Ul a) (dplus a) (neg a)
    (zero a)(one a) (dmult a)(inv a)).
Proof. intros. destruct H.
  set (b:=Fnotation.create  (Ul a) (dplus a) (neg a) (zero a)(one a)
    (dmult a)(inv a)).
  assert (strong_sfield_axioms b). unfold b. app sfield_is_sfield. red.
  destruct H1.
  split. am. split. am.
  assert (Ul b = Ul a). unfold b. mw.
  assert (dmult b = dmult a). unfold b. app mult_fcreate. rw H3; rww H4.
Qed.

Lemma sfield_is_ring: forall a, sfield_axioms a -> ring_axioms a.
Proof. ir. red in H. red in H. destruct H. am.
Qed.

Lemma strong_sfield_is_ring: forall a, sfield_axioms a -> 
  strong_ring_axioms (Rnotation.create (Ul a) (dplus a) (neg a)
    (zero a)(one a)(dmult a)).
Proof. ir. destruct H.
  pose (b:=Rnotation.create (Ul a) (dplus a) (neg a) (zero a)(one a)(dmult a)).
  app rcreate_axioms. app is_bin_domain_plus. app is_un_domain_neg. 
  app is_bin_domain_mult. 
Qed. 

Lemma field_is_cring: forall a, field_axioms a -> cring_axioms a.
Proof. ir. red in H. red in H. destruct H. destruct H. split;am.
Qed.

Lemma strong_field_is_cring: forall a, field_axioms a -> 
  strong_cring_axioms (Rnotation.create (Ul a) (dplus a) (neg a)
    (zero a)(one a)(dmult a)).
Proof. ir. cp (field_is_cring H). cp(cring_is_ring H0). 
  destruct H.
  pose (b:=Rnotation.create (Ul a) (dplus a) (neg a) (zero a)(one a)(dmult a)).
  assert (strong_ring_axioms b). unfold b. app strong_sfield_is_ring.
  red in H3. ee. red in H4. ee. fold b. split. am. split. am.
  assert (Ul b = Ul a). unfold b. mw.
  assert (dmult b = dmult a). unfold b. app mult_rcreate.
  rw H5; rww H6.
Qed. 


Lemma lmcreate_axioms: forall x y p o z m,
  Operations.is_binary_domain p x -> Operations.is_unary_domain o x ->
  Operations.is_action_domain m (Ul y) x -> 
  lmproperty x y p o z m -> strong_ring_axioms y ->
  strong_lmodule_axioms (LMnotation.create x y p o z m).
Proof. ir. assert (ring_axioms y). red in H3; ee;am.
  cp (lmodule_implies_mstruct H2). red. split. mw.
  app LMnotation.like_create. split. mw.  red in H3; ee; am. red. split. mw. mw.
Qed.

Lemma rmcreate_axioms: forall x y p o z m,
  Operations.is_binary_domain p x -> Operations.is_unary_domain o x ->
  Operations.is_action_domain m (Ul y) x -> 
  rmproperty x y p o z m -> strong_ring_axioms y ->
  strong_rmodule_axioms (RMnotation.create x y p o z m).
Proof. ir. assert (ring_axioms y). red in H3; ee;am.
  cp (rmodule_implies_mstruct H2). red. split.
  app RMnotation.like_create. split. mw. red in H3;ee;am. red. split. mw. mw.
Qed.

Lemma lrmcreate_axioms: forall x yr yl p o z ml mr,
  Operations.is_binary_domain p x -> Operations.is_unary_domain o x ->
  Operations.is_action_domain ml (Ul yl) x -> 
  Operations.is_action_domain mr (Ul yr) x -> 
  lrmproperty x yl yr p o z ml mr -> strong_ring_axioms yl ->
  strong_ring_axioms yr ->
  strong_lrmodule_axioms (LRMnotation.create x yl yr p o z ml mr).
Proof. ir. 
  assert (ring_axioms yr). red in H5; ee;am.
  assert (ring_axioms yl). red in H4; ee;am.
  cp (lrmodule_implies_mstruct H3). red. split. 
  app LRMnotation.like_create. split. mw. red in H4; ee;am.
  split. mw. red in H5; ee;am. red. split. mw. mw. split; am.
Qed.


Lemma lmodule_is_lmodule: forall a, lmodule_axioms a -> 
  strong_ring_axioms (lauxiliary a) ->
  strong_lmodule_axioms (LMnotation.create (Ul a) (lauxiliary a) 
    (dplus a) (neg a) (zero a) (lmult a)).
Proof. ir. red in H. ee. ap lmcreate_axioms. app is_bin_domain_plus.
  app is_un_domain_neg. app is_act_domain_lmult.  am. am.
Qed.

Lemma rmodule_is_rmodule: forall a, rmodule_axioms a ->
  strong_ring_axioms (rauxiliary a) ->
  strong_rmodule_axioms (RMnotation.create (Ul a) (rauxiliary a) (dplus a) 
    (neg a) (zero a) (rmult a)).
Proof. ir. red in H. ee. ap rmcreate_axioms. app is_bin_domain_plus.
  app is_un_domain_neg. app is_act_domain_rmult.  am. am.
Qed.

Lemma lrmodule_is_lrmodule: forall a, lrmodule_axioms a ->
  strong_ring_axioms (lauxiliary a) ->
  strong_ring_axioms (rauxiliary a) ->
  strong_lrmodule_axioms (LRMnotation.create (Ul a) (lauxiliary a)  
    (rauxiliary a) (dplus a) (neg a) (zero a) (lmult a) (rmult a)).
Proof. ir. red in H. ee. ap lrmcreate_axioms. app is_bin_domain_plus.
  app is_un_domain_neg. app is_act_domain_lmult. app is_act_domain_rmult.  
  am. am. am.
Qed.

Lemma lmodule_is_cgroup: forall a, lmodule_axioms a -> cgroup_axioms a.
Proof. ir. destruct H as [_ [H _]]. am. 
Qed.

Lemma rmodule_is_cgroup: forall a, rmodule_axioms a -> cgroup_axioms a.
Proof. ir. destruct H as [_ [H _]]. am. 
Qed.

Lemma lrmodule_is_cgroup: forall a, lrmodule_axioms a -> cgroup_axioms a.
Proof. ir. destruct H as [_ [_ [H _]]]. am. 
Qed.

Lemma strong_lmodule_is_cgroup: forall a, lmodule_axioms a ->  
  strong_cgroup_axioms (Gnotation.create (Ul a) (dplus a) (neg a) (zero a)).
Proof. ir. destruct H as [_ [H _]]. destruct H.
 pose (b:=Gnotation.create (Ul a) (dplus a) (neg a) (zero a)).
  app cgcreate_axioms. app is_bin_domain_plus. app is_un_domain_neg. 
Qed. 

Lemma strong_rmodule_is_cgroup: forall a, rmodule_axioms a ->  
  strong_cgroup_axioms (Gnotation.create (Ul a) (dplus a) (neg a) (zero a)).
Proof. ir. destruct H as [_ [H _]]. destruct H.
 pose (b:=Gnotation.create (Ul a) (dplus a) (neg a) (zero a)).
  app cgcreate_axioms. app is_bin_domain_plus. app is_un_domain_neg. 
Qed. 

Lemma strong_lrmodule_is_cgroup: forall a, lrmodule_axioms a ->  
  strong_cgroup_axioms (Gnotation.create (Ul a) (dplus a) (neg a) (zero a)).
Proof. ir. destruct H as [_ [_ [H _]]]. destruct H.
 pose (b:=Gnotation.create (Ul a) (dplus a) (neg a) (zero a)).
  app cgcreate_axioms. app is_bin_domain_plus. app is_un_domain_neg. 
Qed. 

Lemma lmodule_aux_is_ring: forall a, lmodule_axioms a -> 
  ring_axioms (lauxiliary a).
Proof. ir. red in H; ee. am. Qed.
Lemma rmodule_aux_is_ring: forall a, rmodule_axioms a -> 
  ring_axioms (rauxiliary a).
Proof. ir. red in H; ee. am. Qed.
Lemma llrmodule_aux_is_ring: forall a, lrmodule_axioms a -> 
  ring_axioms (lauxiliary a).
Proof. ir. red in H; ee. am. Qed.
Lemma rlrmodule_aux_is_ring: forall a, lrmodule_axioms a -> 
  ring_axioms (rauxiliary a).
Proof. ir. red in H; ee. am. Qed.

Lemma lrmodule_is_lmodule: forall a, lrmodule_axioms a -> 
  lmodule_axioms a. 
Proof. ir.  red in H. ee. red; split. am.  red in H1. ee.
  red.  intuition. Qed.

Lemma lrmodule_is_rmodule: forall a, lrmodule_axioms a -> 
  rmodule_axioms a. 
Proof. ir.  red in H. ee. red; split. am.  red in H1. ee.
  red.  intuition. Qed.

End Definitions.
Export Definitions.

